"""Create hoc files"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import sys
import os
import argparse
import multiprocessing
import csv


from bluepymm import tools, prepare_combos


def get_parser():
    """Return the argument parser"""
    parser = argparse.ArgumentParser(description='Create legacy .hoc files')
    parser.add_argument('conf_filename')

    return parser


def add_full_paths(config, directory):
    """Add full paths based on given directory to values of given config if the
    resulting path is valid.

    Args:
        config: dictionary
        directory: string used to complete paths

    Returns:
        The dictionary with completed paths.
    """
    for k, v in config.items():
        if isinstance(v, str):
            test_path = os.path.join(directory, v)
            print(test_path)
            if os.path.isdir(test_path) or os.path.isfile(test_path):
                config[k] = test_path
    return config


def load_combinations_dict(mecombo_emodel_path):
    """Load combinations dictionary.

    Args:
        mecombo_emodel_path: path to file with me-combo data

    Returns:
        A dictionary
    """
    with open(mecombo_emodel_path) as f:
        reader = csv.DictReader(f, delimiter='\t')
        combinations_dict = {row['combo_name']: row for row in reader}
    return combinations_dict


def run_create_and_write_hoc_file(emodel, setup_dir, hoc_dir, emodel_params,
                                  template, template_dir, morph_path,
                                  model_name):
    """Run create_and_write_hoc_file in isolated environment.

    Args:
        See create_and_write_hoc_file.
    """
    pool = multiprocessing.pool.Pool(1, maxtasksperchild=1)
    pool.apply(prepare_combos.prepare_emodel_dirs.create_and_write_hoc_file,
               (emodel, setup_dir, hoc_dir, emodel_params, template,
                template_dir, morph_path, model_name))
    pool.terminate()
    pool.join()
    del pool


def create_hoc_files(combinations_dict, emodels_dir, final_dict, template,
                     hoc_dir):
    """Create a .hoc file for every combination in a given database.

    Args:
        combinations_dict: Dictionary with e-model - morphology combinations.
        emodels_dir: Directory containing all e-model data as used by the
                     application 'bluepymm'.
        final_dict: Dictionary with e-model parameters.
        template: Template to be used to create .hoc files.
        hoc_dir: Directory where all create .hoc files will be written.
    """
    for combination, comb_data in combinations_dict.items():
        print('Working on combination {}'.format(combination))
        emodel = comb_data['emodel']
        setup_dir = os.path.join(emodels_dir, emodel)
        morph_path = '{}.asc'.format(comb_data['morph_name'])
        emodel_params = final_dict[emodel]['params']

        run_create_and_write_hoc_file(emodel,
                                      setup_dir,
                                      hoc_dir,
                                      emodel_params,
                                      os.path.basename(template),
                                      os.path.dirname(template),
                                      morph_path,
                                      combination)


def main(arg_list):
    """Main"""

    # parse and process arguments
    args = get_parser().parse_args(arg_list)
    config = tools.load_json(args.conf_filename)
    config_dir = os.path.abspath(os.path.dirname(args.conf_filename))
    config = add_full_paths(config, config_dir)

    # process configuration
    mecombo_emodel_filename = config['mecombo_emodel_filename']
    combinations_dict = load_combinations_dict(mecombo_emodel_filename)
    final_dict = tools.load_json(config['final_json_path'])
    emodels_dir = config['emodels_tmp_dir']

    # create output directory for .hoc files
    tools.makedirs(config['hoc_output_dir'])

    # create hoc files
    create_hoc_files(combinations_dict, emodels_dir, final_dict,
                     config['template'], config['hoc_output_dir'])


if __name__ == '__main__':
    main(sys.argv[1:])



"""Create sqlite database"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Some code based on BrainBuilder and morph repair code"""

# pylint: disable=R0912

import pandas
import re

import lxml
import lxml.etree


def _parse_xml_tree(filename):
    """Read xml tree from file.

    Args:
        filename(str): filename of recipe (XML)

    Returns:
        xml.etree.ElementTree
    """
    parser = lxml.etree.XMLParser(resolve_entities=False)
    return lxml.etree.parse(filename, parser=parser)


def verify_no_zero_percentage(tree_element_list):
    """Verify that none of the elements of a given list have a zero value for
    the field 'percentage'.

    Args:
        tree_element_list(list of xml.etree.ElementTree): list of tree elements
            with 'percentage' field

    Returns:
        True if no percentage of zero is found.

    Raises:
        ValueError: if a percentage of zero is found.
    """
    for element in tree_element_list:
        if float(element.attrib['percentage']) == 0.0:
            raise ValueError('Found a percentage of 0.0 in recipe, script'
                             ' cannot handle this case: tag'
                             ' {}'.format(element.tag))
    return True


def read_recipe_records(recipe_tree):
    """Parse recipe tree and yield (layer, m-type, e-type)-tuples.

    Args:
        recipe_tree: xml.etree.ElementTree.ElementTree or
                     xml.etree.ElementTree.Element

    Yields:
        (layer, m-type, e-type)-tuples
    """
    for layer in recipe_tree.findall('NeuronTypes')[0].getchildren():
        for mtype in layer.getchildren():
            if mtype.tag == "StructuralType":
                for etype in mtype.getchildren():
                    if etype.tag == "ElectroType":
                        verify_no_zero_percentage([layer, mtype, etype])
                        yield (layer.attrib['id'],
                               mtype.attrib['id'],
                               etype.attrib['id'])


def read_mm_recipe(recipe_filename):
    """Read a BBP builder recipe and return a pandas.DataFrame with all
    possible (layer, m-type, e-type)-combinations.

    Args:
        recipe_filename(str): filename of recipe (XML)

    Returns:
        A pandas.DataFrame with fields "layer", "fullmtype", and "etype".
    """
    recipe_tree = _parse_xml_tree(recipe_filename)
    return pandas.DataFrame(read_recipe_records(recipe_tree),
                            columns=["layer", "fullmtype", "etype"])


def read_morph_records(morph_tree):
    """Parse morphology tree and yield (name, fullmtype, mtype, msubtype,
    layer)-tuples.

    Args:
        morph_tree: xml.etree.ElementTree.ElementTree or
                    xml.etree.ElementTree.Element

    Yields:
        (name, fullmtype, mtype, msubtype, layer)-tuples
    """
    for morph in morph_tree.findall('.//morphology'):
        name = morph.findtext('name')
        mtype = morph.findtext('mtype')
        msubtype = morph.findtext('msubtype')
        fullmtype = '%s:%s' % (mtype, msubtype) if msubtype != '' else mtype
        layer = morph.findtext('layer')
        yield (name, fullmtype, mtype, msubtype, layer)


def read_mtype_morph_map(neurondb_filename):
    """Read morphology database and return a pandas.DataFrame with all
    morphology records.

    Args:
        neurondb_filename(str): filename of morphology database (XML)

    Returns:
        A pandas.DataFrame with field "morph_name", "fullmtype", "mtype",
        "submtype", "layer".
    """
    xml_tree = _parse_xml_tree(neurondb_filename)
    column_labels = ["morph_name", "fullmtype", "mtype", "submtype", "layer"]
    return pandas.DataFrame(read_morph_records(xml_tree),
                            columns=column_labels)


def convert_emodel_etype_map(emodel_etype_map, fullmtypes, etypes):
    """Resolve regular expressions in an e-model e-type map and convert the
    result to a pandas.DataFrame. In the absence of the key "etype", "mtype",
    or "morph_name" in the e-model e-type map, the regular expression ".*" is
    assumed.

    Args:
        emodel_etype_map: A dict mapping e-models to a dict with keys
            "mm_recipe" and "layer". Optional additional keys are "etype",
            "mtype", and "morph_name", which may contain regular expressions.
            In absence of these keys, the regular expression ".*" is assumed.
        fullmtypes: A set of unique full m-types
        etypes: A set of unique e-types

    Returns:
        A pandas.DataFrame with fields 'emodel', 'layer', 'fullmtype', 'etype',
        'morph_regex', and 'original_emodel'. Each row corresponds to a unique
        e-model description.
    """
    morph_name_regexs_cache = {}

    def read_records():
        """Read records"""
        for original_emodel, etype_map in emodel_etype_map.items():
            etype_regex = re.compile(etype_map.get('etype', '.*'))
            mtype_regex = re.compile(etype_map.get('mtype', '.*'))

            morph_name_regex = etype_map.get('morph_name', '.*')
            morph_name_regex = morph_name_regexs_cache.setdefault(
                morph_name_regex, re.compile(morph_name_regex))

            emodel = etype_map['mm_recipe']
            for layer in etype_map['layer']:
                for fullmtype in fullmtypes:
                    if mtype_regex.match(fullmtype):
                        for etype in etypes:
                            if etype_regex.match(etype):
                                yield (emodel,
                                       str(layer),
                                       fullmtype,
                                       etype,
                                       morph_name_regex,
                                       original_emodel,)

    columns = ['emodel', 'layer', 'fullmtype', 'etype', 'morph_regex',
               'original_emodel']
    return pandas.DataFrame(read_records(), columns=columns)



"""BluePyMM select_combos sqlite input"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=R0914, C0325, W0640


import pandas
import sqlite3


def read_and_process_sqlite_score_tables(scores_sqlite_filename):
    """Read score and score values tables from score sqlite dabatase.

    Args:
        scores_sqlite_filename: path to sqlite database

    Returns:
        A tuple of two pandas.DataFrames, as loaded from the tables 'scores'
        and 'score_values' respectively, If present, the column 'index' is
        dropped from score_values.

    Raises:
        Exception if the number of rows in the 'scores' table does not equal
        the number of rows in the 'score_values' table.
    """

    print('Reading scores and score values from {} ...'.format(
        scores_sqlite_filename))
    with sqlite3.connect(scores_sqlite_filename) as conn:
        scores = pandas.read_sql('SELECT * FROM scores', conn)
        score_values = pandas.read_sql('SELECT * FROM score_values', conn)

    if len(score_values.index) != len(scores.index):
        raise Exception("Score and score values tables don't have same number"
                        " of entries!")

    # every column should correspond to a score.
    if 'index' in score_values.columns.values:
        score_values.drop(labels=['index'], axis=1, inplace=True)

    return scores, score_values



"""BluePyMM tools"""

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import contextlib
import errno
import imp
import json
import os
import sys
import hashlib
from string import digits


@contextlib.contextmanager
def cd(dir_name):
    """Change directory"""
    old_cwd = os.getcwd()
    os.chdir(dir_name)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def load_json(path):
    """Load json file"""
    with open(path) as f:
        return json.load(f)


def write_json(output_dir, output_name, config):
    """Write json file"""
    path = os.path.join(output_dir, output_name)
    with open(path, 'w') as fd:
        json.dump(config, fd, indent=2, sort_keys=True)
    return path


def makedirs(path):
    """mkdir but don't fail when dir exists"""
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise
    return path


def check_no_null_nan_values(data, description):
    """Check whether a pandas.DataFrame contains neither None nor NaN values.

    Returns:
        bool: True if successful.

    Raises:
        Exception: if `data` contains at least one None or NaN value.
    """
    if data.isnull().values.any():
        raise Exception('{} contains None/NaN values.'.format(description))
    return True


def check_all_combos_have_run(database, description):
    """Verify that all entries of a given database have run.

    Args:
        database: a pandas.DataFrame with a column 'to_run'
        description: string that contains description of database

    Returns:
        True if the value of 'to_run' is False for all rows.

    Raises:
        Exception, if the database contains at least one entry where the value
        of 'to_run' is True.
    """
    if database['to_run'].any():
        raise Exception('At least one me-combination of database "{}" has not'
                        ' been run'.format(description))
    else:
        return True


def load_module(name, path):
    """Try and load module `name` but *only* in `path`

    from https://docs.python.org/2/library/imp.html#examples
    """
    # Fast path: see if the module has already been imported.
    try:
        return sys.modules[name]
    except KeyError:
        pass

    fp, pathname, description = imp.find_module(name, [path])
    try:
        return imp.load_module(name, fp, pathname, description)
    finally:
        if fp:
            fp.close()


def check_compliance_with_neuron(template_name):
    """Verify that a given name is compliant with the rules for a NEURON
    template name: a name should be a non-empty alphanumeric string, and start
    with a letter. Underscores are allowed. The length should not exceed 50
    characters.

    Returns:
        True if compliant, false otherwise.
    """
    max_len = 50
    return (template_name and template_name[0].isalpha() and
            template_name.replace('_', '').isalnum() and
            len(template_name) <= max_len)


def shorten_and_hash_string(label, keep_length=40, hash_length=9):
    """Convert string to a shorter string if required.

    Args:
        label: a string to be converted
        keep_length: length of the original string to keep. Default is 40
            characters.
        hash_length: length of the hash to generate, should not be more then
            20. Default is 9 characters.

    Returns:
        If the length of the original label is shorter than the sum of
        'keep_length' and 'hash_length' plus one the original string is
        returned. Otherwise, a string with structure <partial>_<hash> is
        returned, where <partial> is the first part of the original string
        with length equal to <keep_length> and the last part is a hash of
        'hash_length' characters, based on the original string.

    Raises:
        ValueError, if 'hash_length' exceeds 20.
    """

    if hash_length > 20:
        raise ValueError('Parameter hash_length should not exceed 20, '
                         ' received: {}'.format(hash_length))

    if len(label) <= keep_length + hash_length + 1:
        return label

    hash_string = hashlib.sha1(label.encode('utf-8')).hexdigest()
    return '{}_{}'.format(label[0:keep_length], hash_string[0:hash_length])


def get_neuron_compliant_template_name(name):
    """Get template name that is compliant with NEURON based on given name.

    Args:
        name: string

    Returns:
        If `name' is NEURON-compliant, the same string is return. Otherwise,
        hyphens are replaced by underscores and if appropriate, the string is
        shortened. Leading numbers are removed.
    """
    template_name = name
    if not check_compliance_with_neuron(template_name):
        template_name = template_name.lstrip(digits).replace("-", "_")
        template_name = shorten_and_hash_string(template_name,
                                                keep_length=40,
                                                hash_length=9)
    return template_name



"""Create sqlite database"""

from __future__ import print_function

"""
Copyright (c) 2018, EPFL/Blue Brain Project

 This file is part of BluePyMM <https://github.com/BlueBrain/BluePyMM>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

"""Some Code based on BrainBuilder and morph repair code"""

# pylint: disable=R0914

import os
import json

import pandas
import sqlite3

from bluepymm import tools
from . import parse_files


def check_morphology_existence(morph_name, morph_type, morph_path):
    """Check if a morphology exists based on its path.

    Args:
        morph_name: a string representing the name of the morphology. Used for
            makeing a pretty error string.
        morph_type: a string representing the type of the morphology. Used for
            making a pretty error string.
        morph_path: the path to the morphology file

    Returns:
        True if the morphology exists.

    Raises:
        ValueError: The morphology does not exist.
    """
    if not os.path.isfile(morph_path):
        raise ValueError(
            "{} morphology {} doesn't exist at {}".format(
                morph_type.capitalize(), morph_name, morph_path))
    else:
        return True


def create_exemplar_rows(
        final_dict,
        rep_fullmtype_morph_map,
        emodel_etype_map,
        emodel_dirs,
        rep_morph_dir,
        unrep_morph_dir,
        skip_repaired_exemplar=False):
    """Create exemplar rows.

    Args:
        final_dict: final e-model map
        rep_fullmtype_morph_map: pandas.DataFrame with morphology database
        emodel_etype_map: e-model e-type map
        emodel_dirs: a dict mapping e-models to prepared e-model directories
        rep_morph_dir: directory with repaired morphologies
        skip_repaired_exemplar: indicates whether repaired exemplar should be
            skipped. Default value is False.

    Returns:
        pandas.DataFrame with one row for each exemplar. Keys are 'layer',
        'fullmtype', 'mtype', 'msubtype', 'etype', 'morph_name', 'emodel',
        'original_emodel', 'morph_dir', 'scores', 'opt_scores', 'exception',
        'to_run', 'is_exemplar', 'is_repaired', and 'is_original'.
    """

    exemplar_rows = []

    for original_emodel in emodel_etype_map:
        emodel = emodel_etype_map[original_emodel]['mm_recipe']
        print('Adding exemplar row for e-model %s' % emodel)

        original_emodel_dict = final_dict[original_emodel]

        opt_scores = original_emodel_dict['fitness']

        morph_filename = os.path.basename(original_emodel_dict['morph_path'])
        morph_name, morph_ext = os.path.splitext(morph_filename)

        # Warning: use this_ prefix, next iteration in for loop will pick up
        # previous step
        if unrep_morph_dir is None:
            this_unrep_morph_dir = os.path.dirname(os.path.abspath(
                os.path.join(
                    emodel_dirs[emodel],
                    original_emodel_dict['morph_path'])))
        else:
            this_unrep_morph_dir = unrep_morph_dir
        morph_path = os.path.join(this_unrep_morph_dir, morph_filename)

        check_morphology_existence(morph_filename, 'unrepaired', morph_path)

        if skip_repaired_exemplar:
            fullmtype = None
            mtype = None
            msubtype = None
            # Don't run repaired version
            combos = [(emodel, False, False),
                      (original_emodel, True, False)]
        else:
            morph_info_list = rep_fullmtype_morph_map[
                rep_fullmtype_morph_map['morph_name'] == morph_name].values
            if len(morph_info_list) == 0:
                raise Exception(
                    'Morphology %s for %s e-model not found in morphology '
                    'release' % (morph_name, original_emodel))
            else:
                _, fullmtype, mtype, msubtype, _ = morph_info_list[0]

            morph_path = os.path.join(rep_morph_dir, morph_filename)
            check_morphology_existence(morph_filename, 'repaired', morph_path)
            # Run repaired version
            combos = [(emodel, False, True),
                      (original_emodel, True, True),
                      (emodel, False, False),
                      (original_emodel, True, False)]

        for (stored_emodel, original, repaired) in combos:
            new_row_dict = {
                'layer': None,
                'fullmtype': fullmtype,
                'mtype': mtype,
                'msubtype': msubtype,
                'etype': emodel_etype_map[original_emodel]['etype'],
                'morph_name': morph_name,
                'morph_ext': morph_ext,
                'emodel': stored_emodel,
                'original_emodel': original_emodel,
                'morph_dir':
                rep_morph_dir if repaired else this_unrep_morph_dir,
                'scores': None,
                'opt_scores': json.dumps(opt_scores) if not repaired else None,
                'exception': None,
                'to_run': True,
                'is_exemplar': True,
                'is_repaired': repaired,
                'is_original': original}
            exemplar_rows.append(new_row_dict)

    return pandas.DataFrame(exemplar_rows)


def remove_morph_regex_failures(full_map):
    """Remove all rows where morph_name doesn't match morph_regex.

    Args:
        full_map: pandas.DataFrame with keys 'morph_name' and 'morph_regex'

    Returns:
        The processed pandas.DataFrame, with all rows of the input table where
        'morph_name' matched 'morph_regex'. The column 'morph_regex' is
        removed.
    """
    # Add a new column to store the regex match result
    full_map.insert(len(full_map.columns), 'morph_regex_matches', None)

    def match_morph_regex(row):
        """Check if 'morph_name' matches 'morph_regex'."""
        row['morph_regex_matches'] = \
            bool(row['morph_regex'].match(row['morph_name']))
        return row

    # Check if 'morph_name' matches 'morph_regex'
    full_map = full_map.apply(match_morph_regex, axis=1)

    # Prune all the rows that didn't match
    full_map = full_map[full_map['morph_regex_matches'] == True]  # NOQA

    # Delete obsolete columns and reset index
    del full_map['morph_regex']
    del full_map['morph_regex_matches']
    return full_map.reset_index(drop=True)


def create_mm_sqlite(
        output_filename,
        recipe_filename,
        morph_dir,
        rep_morph_dir,
        unrep_morph_dir,
        original_emodel_etype_map,
        final_dict,
        emodel_dirs,
        skip_repaired_exemplar=False):
    """Create SQLite database with all possible me-combinations.

    Args:
        output_filename
        recipe_filename
        morph_dir: directory with morphology release, contains neuronDB.xml
            file
        original_emodel_etype_map
        final_dict: e-model parameters
        emodel_dirs: prepared e-model directories
        skip_repaired_exemplar: indicates whether repaired exemplar should be
            skipped. Default value is False.
    """
    neurondb_filename = os.path.join(morph_dir, 'neuronDB.xml')
    rep_neurondb_filename = os.path.join(rep_morph_dir, 'neuronDB.xml')

    # Contains layer, fullmtype, etype
    print('Reading recipe at %s' % recipe_filename)
    fullmtype_etype_map = parse_files.read_mm_recipe(recipe_filename)
    tools.check_no_null_nan_values(fullmtype_etype_map,
                                   "the full m-type e-type map")

    # Contains layer, fullmtype, mtype, submtype, morph_name
    print('Reading neuronDB at %s' % neurondb_filename)
    fullmtype_morph_map = parse_files.read_mtype_morph_map(neurondb_filename)
    tools.check_no_null_nan_values(fullmtype_morph_map,
                                   "the full m-type morphology map")

    # Contains layer, fullmtype, mtype, submtype, morph_name
    print(
        'Reading repaired-morphologies neuronDB at %s' %
        rep_neurondb_filename)
    rep_fullmtype_morph_map = parse_files.read_mtype_morph_map(
        rep_neurondb_filename)
    tools.check_no_null_nan_values(rep_fullmtype_morph_map,
                                   "the full m-type morphology map")

    # Contains layer, fullmtype, etype, morph_name
    print('Merging recipe and neuronDB tables')
    morph_fullmtype_etype_map = fullmtype_morph_map.merge(
        fullmtype_etype_map, on=['fullmtype', 'layer'], how='left')
    tools.check_no_null_nan_values(morph_fullmtype_etype_map,
                                   "morph_fullmtype_etype_map")

    fullmtypes = morph_fullmtype_etype_map.fullmtype.unique()
    etypes = morph_fullmtype_etype_map.etype.unique()

    print('Creating emodel etype table')
    # Contains layer, fullmtype, etype, emodel, morph_regex, original_emodel
    emodel_fullmtype_etype_map = parse_files.convert_emodel_etype_map(
        original_emodel_etype_map, fullmtypes, etypes)
    tools.check_no_null_nan_values(emodel_fullmtype_etype_map,
                                   "e-model e-type map")

    print('Creating full table by merging subtables')
    # Contains layer, fullmtype, etype, morph_name, e_model, morph_regex
    full_map = morph_fullmtype_etype_map.merge(
        emodel_fullmtype_etype_map,
        on=['layer', 'etype', 'fullmtype'], how='left')

    null_emodel_rows = full_map[pandas.isnull(full_map['emodel'])]

    if len(null_emodel_rows) > 0:
        raise Exception(
            'No emodels found for the following layer, etype, fullmtype'
            ' combinations: \n%s' %
            null_emodel_rows[['layer', 'etype', 'fullmtype']])

    print('Filtering out morp_names that dont match regex')
    # Contains layer, fullmtype, etype, morph_name, e_model
    full_map = remove_morph_regex_failures(full_map)
    tools.check_no_null_nan_values(full_map, "the full map")

    print('Adding exemplar rows')
    full_map.insert(len(full_map.columns), 'morph_dir', morph_dir)
    full_map.insert(len(full_map.columns), 'morph_ext', None)
    full_map.insert(len(full_map.columns), 'is_exemplar', False)
    full_map.insert(len(full_map.columns), 'is_repaired', True)
    full_map.insert(len(full_map.columns), 'is_original', False)
    full_map.insert(len(full_map.columns), 'scores', None)
    full_map.insert(len(full_map.columns), 'opt_scores', None)
    full_map.insert(len(full_map.columns), 'extra_values', None)
    full_map.insert(len(full_map.columns), 'exception', None)
    full_map.insert(len(full_map.columns), 'to_run', True)

    exemplar_rows = create_exemplar_rows(
        final_dict,
        rep_fullmtype_morph_map,
        original_emodel_etype_map,
        emodel_dirs,
        rep_morph_dir,
        unrep_morph_dir,
        skip_repaired_exemplar=skip_repaired_exemplar)

    # Prepend exemplar rows to full_map
    full_map = pandas.concat([exemplar_rows, full_map], ignore_index=True)

    # Write full table to sqlite database
    with sqlite3.connect(output_filename) as conn:
        full_map.to_sql('scores', conn, if_exists='replace')

    print('Created sqlite db at %s' % output_filename)
